@@ -1,14 +1,16 @@
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 from .validate import (
   ValidationError,
   ValidationWarning,
+  validating,
   valid_type,
   valid_keys,
   mapget,
   as_list )
 
 from .norms import (
+  marker_evaluated,
   scalar,
   scalar_list,
   empty_str,
@@ -58,6 +60,13 @@ from .path import (
   combine_ignore_patterns,
   contains )
 
+from .template import (
+  Template,
+  Namespace,
+  template_substitute,
+  TemplateError,
+  NamespaceError)
+
 from .dist_file import (
   dist_base,
   dist_zip,
@@ -65,6 +74,7 @@ from .dist_file import (
   dist_source_dummy,
   dist_source_targz,
   dist_binary_wheel,
+  dist_binary_editable,
   FileOutsideRootError,
   dist_iter,
   dist_copy )
@@ -78,6 +88,9 @@ from .pkginfo import (
 from .builder import (
   Builder )
 
-from .pyproj import (
+from .load_module import (
   EntryPointError,
+  EntryPoint )
+
+from .pyproj import (
   PyProjBase )
