@@ -1,63 +1,63 @@
+from __future__ import annotations
 import os
-import os.path as osp
+from collections.abc import Callable
 from pathlib import (
   Path,
-  PurePath,
   PurePosixPath)
-import io
-import stat
-import logging
+from logging import (
+  getLogger,
+  Logger)
 import re
 from abc import(
   ABC,
   abstractmethod )
-
 from ..norms import (
+  norm_path,
   hash_sha256 )
-
 from ..validate import (
-  ValidationError,
-  FileOutsideRootError,
-  validating )
+  validating,
+  ValidationError)
+from ..path import (
+  resolve)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class dist_base( ABC ):
   """Builder for file distribution
 
   Parameters
   ----------
-  outname : str
+  outname:
     Name of output file.
-  outdir : str | pathlib.Path
+  outdir:
     Path to directory where the file should be copied after completing build.
-  tmpdir : None | str | pathlib.Path
+  tmpdir:
     If not None, uses the given directory to place the temporary file(s) before
     copying to final location.
     May be the same as outdir.
-  named_dirs : None | Dict[ str, str | pathlib.PurePosixPath ]
+  named_dirs:
     Mapping of specially named directories within the distribution.
     By default a named directory { 'root' : '.' } will be added,
-    unless overridden with another directory name.
-  logger : None | logging.Logger
+    unless overridden with another directory name. Must be in form of Posix path.
+  logger:
     Logger to which any status information is to be logged.
 
   Attributes
   ----------
-  outpath : pathlib.Path
+  outpath :
     Path to final output file location
-  named_dirs : Dict[ str, pathlib.PurePosixPath ]
+  named_dirs :
     Mapping of specially named directories within the distribution
-  opened : bool
+  opened :
     Build temporary file has been opened for writing
-  finalized : bool
+  finalized :
     Build temporary file has been finalized.
-  closed : bool
+  closed :
     Build temporary file has been closed
-  copied : bool
+  copied :
     Build temporary has been copied to ``outpath`` location
-  records : List[ Tuple[ pathlib.PurePosixPath, str, int ] ]
-    Recorded list of path, hash, and size (bytes) of files added to distribution
-  record_hash : None | str
+  records :
+    Recorded mapping of path to hash, and size (bytes) of files added to distribution
+  record_hash :
     Final hash value of the record after being finalized
 
     .. note::
@@ -65,16 +65,26 @@ class dist_base( ABC ):
       Not all distribution implementations will create a hash of the record
 
   """
+
+  outpath : Path
+  named_dirs : dict[str, PurePosixPath]
+  opened : bool
+  finalized : bool
+  closed : bool
+  copied : bool
+  records : dict[PurePosixPath, tuple[str, int]]
+  record_hash : str|None
+
   #-----------------------------------------------------------------------------
   def __init__( self,
-    outname,
-    outdir = None,
-    tmpdir = None,
-    logger = None,
-    named_dirs = None ):
+    outname: str,
+    outdir: Path|None = None,
+    tmpdir: Path|None = None,
+    logger: Logger|None = None,
+    named_dirs: dict[str, PurePosixPath]|None = None ):
 
     if logger is None:
-      logger = logging.getLogger( type(self).__name__ )
+      logger = getLogger(type(self).__name__)
 
     if named_dirs is None:
       named_dirs = dict()
@@ -90,8 +100,7 @@ class dist_base( ABC ):
       'root' : PurePosixPath(),
       **named_dirs }
 
-
-    self.records = list()
+    self.records = dict()
     self.record_bytes = None
     self.record_hash = None
 
@@ -117,18 +126,19 @@ class dist_base( ABC ):
 
   #-----------------------------------------------------------------------------
   def write( self,
-    dst,
-    data,
-    mode = None,
-    record = True ):
+    dst: PurePosixPath,
+    data: bytes,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
     """Write data into the distribution file
 
     Parameters
     ----------
-    dst : str | PurePosixPath
-    data : bytes
-    mode : int
-    record : bool
+    dst :
+    data :
+    mode :
+    record :
       Add file to the record
 
     """
@@ -136,22 +146,49 @@ class dist_base( ABC ):
     if record:
       self.record(
         dst = dst,
-        data = data )
+        data = data,
+        exist_ok = exist_ok)
+
+  #-----------------------------------------------------------------------------
+  def write_link( self,
+    dst: PurePosixPath,
+    target: PurePosixPath,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
+    """Write symlink into the distribution file
+
+    Parameters
+    ----------
+    dst :
+    target :
+    mode :
+    record :
+      Add file to the record
+
+    """
+    self.logger.debug(f"write_link {dst} ({target})")
+
+    if record:
+      self.record(
+        dst = dst,
+        data = norm_path(target, parent_ok = True).encode('utf-8'),
+        exist_ok = exist_ok)
 
   #-----------------------------------------------------------------------------
   def makedirs( self,
-    dst,
-    mode = None,
-    exist_ok = False,
-    record = True ):
+    dst: PurePosixPath,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True ):
     """Behaviour similar to os.makedirs into the distribution file
 
     Parameters
     ----------
-    dst : str | PurePosixPath
-    mode : int
-    exist_ok : bool
-    record : bool
+    dst :
+    mode :
+    exist_ok :
+    record :
       Add file to the record
 
     Note
@@ -167,20 +204,20 @@ class dist_base( ABC ):
 
   #-----------------------------------------------------------------------------
   def copyfile( self,
-    src,
-    dst,
-    mode = None,
-    exist_ok = False,
-    record = True ):
-    """Behaviour similar to shutil.copyfile into the distribution file
+    src: Path,
+    dst: PurePosixPath,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True ):
+    """Behavior similar to shutil.copyfile into the distribution file
 
     Parameters
     ----------
-    src : str | pathlib.Path
-    dst : str | pathlib.PurePosixPath
-    mode : int
-    exist_ok : bool
-    record : bool
+    src :
+    dst :
+    mode :
+    exist_ok :
+    record :
       Add file to the RECORD
     """
     src = Path(src)
@@ -189,10 +226,7 @@ class dist_base( ABC ):
     if not src.exists():
       raise ValueError(f"Source file not found: {src}")
 
-    if not exist_ok and self.exists( dst ):
-      raise ValueError(f"Build file already has entry: {dst}")
-
-    self.logger.debug( f'copyfile {src}' )
+    self.logger.debug(f'copyfile {src} -> {dst}')
 
     if mode is None:
       mode = src.stat().st_mode
@@ -202,25 +236,25 @@ class dist_base( ABC ):
         dst = dst,
         data = fp,
         mode = mode,
-        record = record )
+        record = record,
+        exist_ok = exist_ok)
 
     return dst
 
   #-----------------------------------------------------------------------------
   def copytree( self,
-    src,
-    dst,
-    ignore = None,
-    exist_ok = False,
-    record = True ):
-    """Behaviour similar to shutil.copytree into the distribution file
+    src: Path,
+    dst: PurePosixPath,
+    ignore: Callable[[Path, list[str]], list[str]]|None = None,
+    exist_ok: bool = False,
+    record: bool = True ):
+    """Behavior similar to shutil.copytree into the distribution file
 
     Parameters
     ----------
-    src : str | pathlib.Path
-    dst : str | pathlib.PurePosixPath
-    ignore : None | callable
-
+    src :
+    dst :
+    ignore :
       If not None, ``callable(src, names) -> ignored_names``
 
       See :func:`shutil.copytree`
@@ -231,11 +265,10 @@ class dist_base( ABC ):
     src = Path(src)
     dst = PurePosixPath(dst)
 
-
     if not src.exists():
       raise ValueError(f"Source directory not found: {src}")
 
-    self.logger.debug( f'copytree {src}' )
+    self.logger.debug(f'copytree {src} -> {dst}')
 
     # returns an iterator of DirEntry
     entries = list(os.scandir(src))
@@ -259,7 +292,20 @@ class dist_base( ABC ):
       mode = entry.stat().st_mode
 
       with validating(key = re.sub( r"[^\w\d]+", "_", entry.name )):
-        if entry.is_dir():
+
+        if entry.is_symlink():
+
+          target = Path(os.readlink(src_path))
+
+          # if not target.is_absolute():
+          #   # ensure minimal path within distribution
+          #   target = resolve(src_path/target)
+
+          # target = target.relative_to(src_path)
+
+          self.write_link(dst_path, target, mode=mode)
+
+        elif entry.is_dir():
 
           self.makedirs(
             dst = dst_path,
@@ -317,50 +363,67 @@ class dist_base( ABC ):
 
   #-----------------------------------------------------------------------------
   def record( self,
-    dst,
-    data ):
+    dst: PurePosixPath,
+    data: bytes,
+    exist_ok: bool = False) -> tuple[str, int]|None:
     """Creates a record for an added file
 
     This produces an sha256 hash of the data and associates a record with the item
 
     Parameters
     ----------
-    dst : str | PurePosixPath
+    dst:
       Path of item within the distribution
-    data : bytes
+    data:
       Binary data that was added
+    exist_ok:
+      Allow overwriting a record. If False, this does *not* raise an exception if
+      the new record would be exactly the same as the previous one for the same
+      archive path.
 
     Returns
     -------
-    None
+    The record of the file `(hash, size)`, or None if the file has already been
+    recorded.
     """
 
     self.assert_recordable()
+    dst = PurePosixPath(dst)
+    _record = self.records.get(dst)
 
-    hash, size = hash_sha256( data )
+    hash, size = hash_sha256(data)
+    record = (hash, size)
 
-    record = ( PurePosixPath(dst), hash, size )
+    if _record is not None:
+      if record == _record:
+        # NOTE: allow equivalent records, leave original
+        return None
 
-    self.logger.debug( 'record ' + str(record) )
+      if not exist_ok:
+        raise ValidationError(f"Overwriting destination: {dst}")
 
-    self.records.append( record )
+      self.logger.debug(f'record overwrite {dst}: {record}')
+
+    else:
+      self.logger.debug(f'record {dst}: {record}')
+
+    self.records[dst] = record
+
+    return record
 
   #-----------------------------------------------------------------------------
   def close( self,
-    finalize = True,
-    copy = True ):
+    finalize: bool = True,
+    copy: bool = True ):
     """Closes the temporary distribution file
 
     Parameters
     ----------
-    finalize : bool
+    finalize :
       If true, finalizes the temporary distribution file before closing
-    copy : bool
+    copy :
       If true, copies the temporary file to final location after closing
 
-    Returns
-    -------
-    None
     """
 
     if self.closed:
@@ -441,12 +504,12 @@ class dist_base( ABC ):
 
   #-----------------------------------------------------------------------------
   @abstractmethod
-  def finalize( self ):
+  def finalize( self ) -> str|None:
     """Finalizes the distribution file before being closed
 
     Returns
     -------
-    record_hash : None | str
+    record_hash :
       sha256 hash of the record
 
     Note
