@@ -1,25 +1,19 @@
+from __future__ import annotations
 import os
-import os.path as osp
 from pathlib import (
   Path,
-  PurePath,
   PurePosixPath)
 import io
-import warnings
-import stat
-
 import tempfile
 import shutil
 import tarfile
-
 from .dist_base import dist_base
-
 from ..norms import (
   norm_path,
   norm_data,
   norm_mode )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class dist_targz( dist_base ):
   """Builds a tar-file  with gz compression
 
@@ -120,7 +114,7 @@ class dist_targz( dist_base ):
       # NOTE: the missing_ok parameter was not added until py38
       self.outpath.unlink()
 
-    self.outdir.mkdir( parents = True, exist_ok = True )
+    self.outdir.mkdir(parents = True, exist_ok = True)
     shutil.copyfile( self._tmp_path, self.outpath )
 
   #-----------------------------------------------------------------------------
@@ -131,23 +125,37 @@ class dist_targz( dist_base ):
     # remove temporary file
     if self._tmp_path.exists():
       self._tmp_path.unlink()
-      
+
     self._tmp_path = None
 
   #-----------------------------------------------------------------------------
   def write( self,
-    dst,
-    data,
-    mode = None,
-    record = True ):
+    dst: PurePosixPath,
+    data: bytes,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
 
     self.assert_open()
 
-    dst = norm_path( os.fspath(dst) )
+    dst = norm_path(dst)
+    data = norm_data(data)
 
-    data = norm_data( data )
+    if record:
+      rec = self.record(
+        dst = dst,
+        data = data,
+        exist_ok = exist_ok)
 
-    info = tarfile.TarInfo( dst )
+      if rec is None:
+        # equivalent file has already been added
+        return
+
+    elif not exist_ok and self.exists( dst ):
+      # NOTE: can only skip equivalent files when they are recorded
+      raise ValueError(f"Overwriting destination: {dst}")
+
+    info = tarfile.TarInfo(dst)
 
     info.size = len(data)
     info.mode = norm_mode( mode )
@@ -156,11 +164,39 @@ class dist_targz( dist_base ):
       info,
       fileobj = io.BytesIO(data) )
 
-    super().write(
-      dst = dst,
-      data = data,
-      mode = mode,
-      record = record )
+  #-----------------------------------------------------------------------------
+  def write_link( self,
+    dst: PurePosixPath,
+    target: PurePosixPath,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
+
+    self.assert_open()
+    dst = norm_path(dst)
+    target = norm_path(target, parent_ok = True)
+    self.logger.debug(f"write_link {dst} ({target})")
+
+    data = target.encode('utf-8')
+
+    if record:
+      rec = self.record(
+        dst = dst,
+        data = data,
+        exist_ok = exist_ok)
+
+      if rec is None:
+        # equivalent file has already been added
+        return
+
+    elif not exist_ok and self.exists( dst ):
+      # NOTE: can only skip equivalent files when they are recorded
+      raise ValueError(f"Overwriting destination: {dst}")
+
+    info = tarfile.TarInfo(dst)
+    info.type = tarfile.SYMTYPE
+    info.linkname = target
+    self._tarfile.addfile(info)
 
   #-----------------------------------------------------------------------------
   def finalize( self ): # pragma: no cover
