@@ -1,26 +1,19 @@
+from __future__ import annotations
 import os
-import os.path as osp
 from pathlib import (
   Path,
-  PurePath,
   PurePosixPath)
-import io
-import warnings
-import stat
-
 import tempfile
 import shutil
 import zipfile
-
+import stat
 from .dist_base import dist_base
-
 from ..norms import (
   norm_path,
   norm_data,
-  norm_mode,
   norm_zip_external_attr )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class dist_zip( dist_base ):
   """Builds a zip file
 
@@ -118,7 +111,7 @@ class dist_zip( dist_base ):
       # NOTE: the missing_ok parameter was not added until py38
       self.outpath.unlink()
 
-    self.outdir.mkdir( parents = True, exist_ok = True )
+    self.outdir.mkdir(parents = True, exist_ok = True )
     shutil.copyfile( self._tmp_path, self.outpath )
 
   #-----------------------------------------------------------------------------
@@ -129,15 +122,16 @@ class dist_zip( dist_base ):
     # remove temporary file
     if self._tmp_path.exists():
       self._tmp_path.unlink()
-      
+
     self._tmp_path = None
 
   #-----------------------------------------------------------------------------
   def write( self,
     dst,
     data,
-    mode = None,
-    record = True ):
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
 
     self.assert_open()
 
@@ -145,6 +139,20 @@ class dist_zip( dist_base ):
 
     data = norm_data( data )
 
+    if record:
+      rec = self.record(
+        dst = dst,
+        data = data,
+        exist_ok = exist_ok)
+
+      if rec is None:
+        # equivalent file has already been added
+        return
+
+    elif not exist_ok and self.exists( dst ):
+      # NOTE: can only skip equivalent files when they are recorded
+      raise ValueError(f"Overwriting destination: {dst}")
+
     zinfo = zipfile.ZipInfo( dst )
 
     zinfo.external_attr = norm_zip_external_attr( mode )
@@ -154,11 +162,40 @@ class dist_zip( dist_base ):
       data,
       compress_type = zipfile.ZIP_DEFLATED )
 
-    super().write(
-      dst = dst,
-      data = data,
-      mode = mode,
-      record = record )
+  #-----------------------------------------------------------------------------
+  def write_link( self,
+    dst: PurePosixPath,
+    target: PurePosixPath,
+    mode: int|None = None,
+    exist_ok: bool = False,
+    record: bool = True):
+
+    self.assert_open()
+    dst = norm_path(dst)
+    target = norm_path(target, parent_ok = True)
+    self.logger.debug(f"write_link {dst} ({target})")
+
+    data = target.encode('utf-8')
+
+    if record:
+      rec = self.record(
+        dst = dst,
+        data = data,
+        exist_ok = exist_ok)
+
+      if rec is None:
+        # equivalent file has already been added
+        return
+
+    elif not exist_ok and self.exists( dst ):
+      # NOTE: can only skip equivalent files when they are recorded
+      raise ValueError(f"Overwriting destination: {dst}")
+
+    zinfo = zipfile.ZipInfo(dst)
+    zinfo.external_attr = norm_zip_external_attr(mode, islink = True)
+
+    self._zipfile.writestr(zinfo, data, compress_type = zipfile.ZIP_DEFLATED)
+
 
   #-----------------------------------------------------------------------------
   def finalize( self ): # pragma: no cover
