@@ -1,3 +1,4 @@
+from __future__ import annotations
 import os
 import os.path as osp
 from pathlib import PurePath
@@ -7,7 +8,7 @@ import logging
 import tempfile
 
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def legacy_setup_content( pyproj, sdist ):
   """
 
@@ -34,7 +35,7 @@ def legacy_setup_content( pyproj, sdist ):
 
   sources = '\n'.join( [
     os.fspath(file)
-    for file, hash, size in sdist.records ] ).encode('utf-8')
+    for file, (hash, size) in sdist.records.items() ] ).encode('utf-8')
 
   with open( osp.join( osp.dirname(__file__), '_legacy_setup.py' ), 'r' ) as fp:
     legacy_setup_py = fp.read()
