@@ -1,12 +1,11 @@
-import sys
+from __future__ import annotations
 import os
-import os.path as osp
-import io
-import warnings
+from io import (
+  IOBase,
+  BytesIO)
 import stat
 import re
 import pathlib
-import inspect
 from copy import copy
 from collections.abc import (
   Mapping,
@@ -20,7 +19,6 @@ from email.message import Message
 from email.generator import BytesGenerator
 from email.utils import parseaddr, formataddr
 from urllib.parse import urlparse
-import keyword
 
 from packaging.tags import sys_tags
 from packaging.markers import Marker
@@ -35,7 +33,7 @@ from .validate import (
   mapget,
   as_list)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def marker_evaluated(marker):
   if isinstance(marker, bool):
     return marker
@@ -44,7 +42,7 @@ def marker_evaluated(marker):
 
   return marker.evaluate()
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def scalar(val):
   if isinstance(val, str):
     return val
@@ -54,11 +52,11 @@ def scalar(val):
 
   return val
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class scalar_list(valid_list):
-  _value_valid = valid(scalar)
+  value_valid = valid(scalar)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_bool(val):
   t = [True, 'true', 'True', 'yes', 'y', 'enable', 'enabled']
   f = [False, 'false', 'False', 'no', 'n', 'disable', 'disabled']
@@ -71,7 +69,7 @@ def norm_bool(val):
 
   return val
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def empty_str(val):
   val = str(val)
 
@@ -81,7 +79,7 @@ def empty_str(val):
 
   return val
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def nonempty_str(val):
   val = str(val)
 
@@ -91,43 +89,44 @@ def nonempty_str(val):
 
   return val
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class str_list(valid_list):
-  _value_valid = valid(str)
+  value_valid = valid(str)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class nonempty_str_list(valid_list):
-  _value_valid = valid(nonempty_str)
+  value_valid = valid(nonempty_str)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def norm_data( data ):
+#===============================================================================
+def norm_data(data: str|bytes|IOBase) -> bytes:
   """Normalize data for writing into a distribution
 
   Parameters
   ----------
-  data : bytes | str | io.IOBase
+  data:
+    Data to process
 
   Returns
   -------
-  data : bytes
+  data:
 
     * If data is bytes, it will be returned un-modified.
     * If data is a str, it will be encoded as 'utf-8'.
     * If data is a stream, it will be read to EOF and apply one of the above.
   """
 
-  if isinstance( data, io.IOBase ):
+  if isinstance(data, IOBase):
     # read from a stream, assumes readable
     data = data.read()
 
-  if isinstance( data, str ):
+  if isinstance(data, str):
     # encode text as utf-8
     data = data.encode('utf-8')
 
   return data
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def norm_path( path ):
+#===============================================================================
+def norm_path( path: str|pathlib.PurePath, parent_ok: bool = False) -> str:
   """Normalizes a file path for writing into a distribution archive
 
   Note
@@ -138,7 +137,7 @@ def norm_path( path ):
   * All slashes replaced with forward slashes
   """
 
-  path = str(path)
+  path = os.fspath(path)
 
   if re.search( r'\s+', path ):
     raise ValidationError(
@@ -151,8 +150,12 @@ def norm_path( path ):
   wpath = pathlib.PureWindowsPath( path )
   path = wpath.as_posix()
 
+  if not parent_ok and re.search(r'(\.\.)', path):
+    raise ValidationError(
+      f"path segments can not be relative to parent directories: {path}")
+
   for segment in path.split('/'):
-    if windows_invalid_filename.fullmatch( segment ):
+    if segment != '..' and windows_invalid_filename.fullmatch(segment):
       raise ValidationError(
         f"Path segment '{segment}' invalid on Windows platform: {path}")
 
@@ -161,14 +164,9 @@ def norm_path( path ):
   if wpath.is_absolute() or ppath.is_absolute():
     raise ValidationError(f"path must be relative: {path}")
 
-
-  if re.search( r'(\.\.)', path ):
-    raise ValidationError(
-      f"path segments can not be relative to parent directories: {path}")
-
   return path
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_path_to_os( path ):
   """Converts a normalized or OS path to be an OS path
   """
@@ -184,7 +182,7 @@ def norm_path_to_os( path ):
 
   return str(pathlib.PurePath(path))
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_mode( mode = None ):
   """Normalizes file permission mode for distribution archive
 
@@ -233,8 +231,8 @@ def norm_mode( mode = None ):
 
   return _mode
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def norm_zip_external_attr( mode = None ):
+#===============================================================================
+def norm_zip_external_attr(mode = None, islink: bool = False):
   """Converts the unix integer mode to zip external_attr
 
   The returned value follows the 4 byte format
@@ -247,6 +245,8 @@ def norm_zip_external_attr( mode = None ):
   Parameters
   ----------
   mode : int
+  islink: bool
+    Set the link flag
 
   Returns
   -------
@@ -258,6 +258,10 @@ def norm_zip_external_attr( mode = None ):
 
   # mask and shift to 3rd byte
   xattr &= 0xFFFF
+
+  if islink:
+    xattr |= stat.S_IFLNK
+
   xattr = xattr << 16
 
   # set directory flag after shift
@@ -269,36 +273,27 @@ def norm_zip_external_attr( mode = None ):
 
   return xattr
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def b64_nopad( data ):
+#===============================================================================
+def b64_nopad(data: bytes) -> str:
   """Encodes hash as urlsafe base64 encoding with no trailing '=' (:pep:`427`)
 
-  Parameters
-  ----------
-  data : bytes
-
-  Returns
-  -------
-  str
-
   See Also
   --------
   * https://www.python.org/dev/peps/pep-0427/#appendix
   """
   return urlsafe_b64encode( data ).decode("ascii").rstrip("=")
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-def hash_sha256( stream ):
+#===============================================================================
+def hash_sha256(stream: BytesIO|bytes) -> tuple[str, int]:
   """Computes SHA-256 hash
 
   Parameters
   ----------
-  stream : bytes | io.BytesIO
+  stream:
 
   Returns
   -------
-  str, int
-    urlsafe base64 encoded hash, and size (in bytes) of the hashed data
+  urlsafe base64 encoded hash, and size (in bytes) of the hashed data
 
   See Also
   --------
@@ -306,7 +301,7 @@ def hash_sha256( stream ):
   """
 
   if isinstance( stream, bytes ):
-    with io.BytesIO( stream ) as _stream:
+    with BytesIO( stream ) as _stream:
       return hash_sha256( _stream )
 
   hasher = hashlib.sha256()
@@ -329,7 +324,7 @@ def hash_sha256( stream ):
 
   return digest_b64_nopad, size
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def email_encode_items(
   headers,
   payload = None ):
@@ -358,7 +353,7 @@ def email_encode_items(
   if payload is not None:
     msg.set_payload( payload )
 
-  buffer = io.BytesIO()
+  buffer = BytesIO()
 
   gen = BytesGenerator(
     buffer,
@@ -373,7 +368,7 @@ def email_encode_items(
 
   return bytes
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class TimeEncode:
   digits = '0123456789abcdefghijklmnopqrstuvwxyz'
 
@@ -387,12 +382,14 @@ class TimeEncode:
     Parameters
     ----------
     resolution : int
-      Number of seconds that are resolved
+      Number of seconds that are resolved.
+      The default resolution is 1 minute
     rollover : int
       Number of seconds upper bound before wrapping.
-      The number of distinct values is ``rollover // resolution``
+      The number of distinct values is ``rollover // resolution``.
+      The default rollover is 324 days
     width : int
-      Minimal witdth of encoded number
+      Minimal width of encoded number
     base : int
       Numeric base used to encode number
     """
@@ -425,7 +422,7 @@ class TimeEncode:
 
     return f'{ts:0>{self.width}}'
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # File/directory names that are invalid on windows
 windows_invalid_filename = '|'.join([
   # filename ending with a period
