@@ -1,48 +1,26 @@
+from __future__ import annotations
 import sys
-import os
-import os.path as osp
-import io
-import warnings
-import stat
 import re
-import pathlib
 import inspect
-from copy import copy
-from collections.abc import (
-  Mapping,
-  Sequence,
-  Iterable )
 
 from collections import namedtuple
-import hashlib
-from base64 import urlsafe_b64encode
-from email.message import Message
-from email.generator import BytesGenerator
 from email.utils import parseaddr, formataddr
 from urllib.parse import urlparse
 import keyword
 
 from packaging.tags import sys_tags
-from packaging.requirements import Requirement
-from packaging.specifiers import SpecifierSet
 
 from .validate import (
   ValidationError,
-  validating,
-  valid_type,
-  valid_keys,
-  valid_dict,
-  valid_list,
-  mapget,
-  as_list)
+  validating)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 CompatibilityTags = namedtuple('CompatibilityTags', ['py_tag', 'abi_tag', 'plat_tag'])
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # NOTE: patterns used for validation are defined at the end of this file
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class PEPValidationError( ValidationError ):
   """Error from value incompatible with a :term:`PEP`
 
@@ -63,7 +41,7 @@ class PEPValidationError( ValidationError ):
     super().__init__(
       msg = f'{msg} (PEP {pep}): {val}' )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_printable(
   text = None ):
   r"""Removes leading and trailing whitespace and all non-printable characters,
@@ -112,7 +90,7 @@ def norm_printable(
 
   return nonprintable.sub( '', str(text).strip() )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def valid_dist_name( name ):
   """Checks for valid distribution name (:pep:`426`)
 
@@ -131,7 +109,7 @@ def valid_dist_name( name ):
 
   return name
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_name( name ):
   """Normalizes a distribution name (:pep:`503`)
 
@@ -150,11 +128,11 @@ def norm_dist_name( name ):
 
   # > The name should be lowercased with all runs of the
   # > characters ., -, or _ replaced with a single - character.
-  name = re.sub( r'[\-\_\.]+', "-", name )
+  name = pep_503_name_norm.sub('-', name)
 
   return name
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_filename( name ):
   """Normalize distribution filename component (:pep:`427`)
 
@@ -163,14 +141,16 @@ def norm_dist_filename( name ):
   Each component of the filename is escaped by replacing runs of
   non-alphanumeric characters with an underscore '_'
 
+  Addendum - It seems that "local" versions require '+'
+
   See Also
   --------
   * https://www.python.org/dev/peps/pep-0427/#file-name-convention
   """
 
-  return re.sub( r"[^\w\d\.]+", "_", name )
+  return re.sub( r"[^\w\d\.\+]+", "_", name )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def join_dist_filename( parts ):
   """Joins distribution filename component (:pep:`427`)
 
@@ -189,10 +169,16 @@ def join_dist_filename( parts ):
     for p in parts
     if p != ''])
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_version( version ):
   """Checks for valid distribution version (:pep:`440`)
 
+  .. versionchanged:: 0.1.9
+
+    Allow local version identifiers ``<public version identifier>[+<local version label>]``,
+    in addition to public versions.
+    Version pattern now uses :ref:`~packaging.version.VERSION_PATTERN`.
+
   See Also
   --------
   * https://www.python.org/dev/peps/pep-0440/#version-scheme
@@ -209,7 +195,7 @@ def norm_dist_version( version ):
 
   return version
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_author(
   name = None,
   email = None ):
@@ -262,7 +248,7 @@ def norm_dist_author(
   # > as appropriate.
   return name, email
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_author_dict(val):
 
   name = norm_printable( val.get('name', '') )
@@ -309,7 +295,7 @@ def norm_dist_author_dict(val):
 
   return val
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_classifier( classifier ):
   """
   See Also
@@ -332,7 +318,7 @@ def norm_dist_classifier( classifier ):
 
   return classifier
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_keyword( keyword ):
   """
   See Also
@@ -350,7 +336,7 @@ def norm_dist_keyword( keyword ):
 
   return keyword
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_url( label, url ):
   """
   See Also
@@ -385,26 +371,40 @@ def norm_dist_url( label, url ):
 
   return label, url
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_extra( extra ):
   """Normalize distribution 'extra' requirement
 
+  .. versionchanged:: 0.2.0
+
+    Extra names are normalized according to PEP-685 and validated according to
+    Core Metadata 2.3.
+    Previously, extra names "must be a valid Python identifier" (Core Metadata 2.1)
+
+
   Note
   ----
-  * No known PEP specifies this format, but is treated as
+  * MUST write out extra names in their normalized form.
+  * This applies to the Provides-Extra field and the extra marker when used
+    in the Requires-Dist field.
+
+  See Also
+  --------
+  * https://peps.python.org/pep-0685/#specification
   """
 
-  extra = norm_printable( extra )
+  extra = norm_printable(extra).lower()
+  extra = pep_503_name_norm.sub('-', extra)
 
-  if not pep_621_extra.fullmatch( extra ):
+  if not pep_685_extra.fullmatch(extra):
     raise PEPValidationError(
-      pep = 621,
+      pep = 685,
       msg = "Invalid extra",
       val = extra )
 
   return extra
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def dist_build( build_number = None, build_tag = None ):
   if build_number is None and build_tag is None:
     build = ''
@@ -420,7 +420,7 @@ def dist_build( build_number = None, build_tag = None ):
 
   return norm_dist_build(build)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_build( build ):
   """
   Note
@@ -443,7 +443,7 @@ def norm_dist_build( build ):
 
   return build
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_dist_compat( py_tag, abi_tag, plat_tag ):
   """
 
@@ -500,7 +500,7 @@ def norm_dist_compat( py_tag, abi_tag, plat_tag ):
 
   return CompatibilityTags( py_tag, abi_tag, plat_tag )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def join_dist_compat( tags ):
   """
   See Also
@@ -509,7 +509,7 @@ def join_dist_compat( tags ):
   """
   return '.'.join( sorted(list(set(tags))) )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def compress_dist_compat( compat ):
   """
   See Also
@@ -525,7 +525,7 @@ def compress_dist_compat( compat ):
 
   return py_tags, abi_tags, plat_tags
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def purelib_compat_tags():
   """Return general compatability tags for the current system
   """
@@ -534,7 +534,7 @@ def purelib_compat_tags():
 
   return compat
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def platlib_compat_tags():
   """Get platform compatability tags for the current system
   """
@@ -547,7 +547,7 @@ def platlib_compat_tags():
 
   return compat_tags
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_py_identifier( name ):
 
   name = norm_printable( name )
@@ -564,7 +564,7 @@ def norm_py_identifier( name ):
 
   return name
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_entry_point_group( name ):
   """Normalizes entry point group
 
@@ -582,13 +582,15 @@ def norm_entry_point_group( name ):
 
   return name
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_entry_point_name( name ):
   """Normalizes entry point name
 
   See Also
   --------
   * https://packaging.python.org/en/latest/specifications/entry-points/
+  * The name may contain any characters except =, but it cannot start or end with
+    any whitespace character, or start with [
   """
 
   name = norm_printable( name )
@@ -600,7 +602,7 @@ def norm_entry_point_name( name ):
 
   return name
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 def norm_entry_point_ref( ref ):
   """Normalizes entry point object reference
 
@@ -633,15 +635,50 @@ def norm_entry_point_ref( ref ):
       msg = f"""Entry point reference must have the form 'importable.module'
         or 'importable.module:object.attr': {ref}""") from e
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
+# https://packaging.python.org/en/latest/specifications/name-normalization/#name-format
 pep426_dist_name = re.compile(
   r'^([A-Z0-9]|[A-Z0-9][A-Z0-9._\-]*[A-Z0-9])$',
   re.IGNORECASE )
 
-pep440_version = re.compile(
-  r'^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*'
-  r'((a|b|rc)(0|[1-9][0-9]*))?'
-  '(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$' )
+# https://packaging.python.org/en/latest/specifications/name-normalization/#name-normalization
+# > runs of characters ., -, or _ replaced with a single - character.
+pep_503_name_norm = re.compile(r'[\-\_\.]+', re.IGNORECASE)
+
+# value of packaging.version.VERSION_PATTERN, as of 'packaging == 25.0'
+# just in case the variable is ever deprecated
+VERSION_PATTERN = r"""
+    v?
+    (?:
+        (?:(?P<epoch>[0-9]+)!)?                           # epoch
+        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
+        (?P<pre>                                          # pre-release
+            [-_\.]?
+            (?P<pre_l>alpha|a|beta|b|preview|pre|c|rc)
+            [-_\.]?
+            (?P<pre_n>[0-9]+)?
+        )?
+        (?P<post>                                         # post release
+            (?:-(?P<post_n1>[0-9]+))
+            |
+            (?:
+                [-_\.]?
+                (?P<post_l>post|rev|r)
+                [-_\.]?
+                (?P<post_n2>[0-9]+)?
+            )
+        )?
+        (?P<dev>                                          # dev release
+            [-_\.]?
+            (?P<dev_l>dev)
+            [-_\.]?
+            (?P<dev_n>[0-9]+)?
+        )?
+    )
+    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
+"""
+
+pep440_version = re.compile(VERSION_PATTERN, re.VERBOSE | re.IGNORECASE)
 
 # NOTE: PEP 427 does not specify any constraints on the string following the
 # digits, but given the form it is used in the filenames it really cannot
@@ -654,7 +691,7 @@ pep425_pytag = re.compile(
   r'^([A-Z0-9_]+)$',
   re.IGNORECASE )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # https://www.python.org/dev/peps/pep-0621/#authors-maintainers
 # https://www.rfc-editor.org/rfc/inline-errata/rfc5322.html
 # > name value MUST be a valid email name (i.e. whatever can be put as a name,
@@ -691,7 +728,7 @@ common_plattag = {
     re.IGNORECASE ),
   'local' : re.compile( rf'^({local_plat})$', re.IGNORECASE )}
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # https://www.python.org/dev/peps/pep-0301/#distutils-trove-classification
 # > It was decided that strings would be used for the classification entries
 # > due to the deep nesting that would be involved in a more formal Python
@@ -707,30 +744,36 @@ pep_301_classifier = re.compile(
   r'^[A-Z0-9._\-\/\[\]\(\) ]+$',
   re.IGNORECASE )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # https://packaging.python.org/en/latest/specifications/core-metadata/#keywords
 # https://www.python.org/dev/peps/pep-0621/#keywords
 # NOTE: does not say what is a valid keyword, but does say they are comma separted,
 # and other implemented with space separated. To be safe ensure no white-space or commas
 pep_621_keyword = re.compile( r'^[^\s\,]+$' )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
-pep_621_extra = re.compile( r'^([A-Z0-9_]+)?$', re.IGNORECASE  )
+#===============================================================================
+# https://packaging.python.org/en/latest/specifications/core-metadata/#core-metadata-provides-extra
+pep_685_extra = re.compile( r'^[a-z0-9]+(-[a-z0-9]+)*$', re.IGNORECASE)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+
+#===============================================================================
 # https://packaging.python.org/en/latest/specifications/entry-points/
 # Group names must be one or more groups of letters, numbers and underscores,
 # separated by dots
 entry_point_group = re.compile( r'^[A-Z0-9_]+(\.[A-Z0-9_]+)*$', re.IGNORECASE  )
+
+# The name may contain any characters except =, but it cannot start or end with
+# any whitespace character, or start with [
 # For new entry points (names), it is recommended to use only letters, numbers,
 # underscores, dots and dashes
-entry_point_name = re.compile( r'^([A-Z0-9_\.\-]+)?$', re.IGNORECASE  )
+# entry_point_name = re.compile(r'^([A-Z0-9_\.\-]+)?$', re.IGNORECASE)
+entry_point_name = re.compile(r'^([^\[\]\=\s]+)?$', re.IGNORECASE)
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 py_keyword = re.compile( '^(' + '|'.join(keyword.kwlist) + ')$' )
 py_identifier = re.compile( r'^[A-Z_][A-Z0-9_]*$', re.IGNORECASE )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 # NOTE: there may be a more efficient way to strip all non-printable characters
 # Here consider new-lines '\n' and tabs '\t' to be printable
 # even though '\n'.isprintable() returns False
