@@ -1,65 +1,53 @@
-import os
-import os.path as osp
-import sys
-import shutil
-import logging
-import tempfile
-from copy import copy, deepcopy
-from collections.abc import (
-  Mapping,
-  Sequence )
+from __future__ import annotations
+from logging import (
+  getLogger,
+  Logger)
+from copy import deepcopy
 import subprocess
-import multiprocessing
-import glob
 import warnings
 import tomli
 from pathlib import (
-  Path,
-  PurePath,
-  PurePosixPath)
-
-try:
-  from importlib.metadata import metadata
-
-except ImportError:
-  from importlib_metadata import metadata
+  Path)
+from importlib import metadata
 
 from .pkginfo import (
   PkgInfoReq,
   PkgInfo )
-
 from .validate import (
   ValidationWarning,
   ValidationError,
   FileOutsideRootError,
+  RequiredValueError,
   valid_dict,
   validating,
   valid,
   restrict,
   mapget )
-
 from .norms import (
   scalar_list,
   norm_bool,
-  norm_path_to_os,
-  norm_path )
-
+  hash_sha256)
 from .pep import (
-  purelib_compat_tags,
   platlib_compat_tags )
-
+from .path import (
+  resolve)
 from .load_module import (
-  EntryPointError,
-  EntryPoint,
-  load_module,
-  load_entrypoint )
+  EntryPoint )
 
 from .legacy import legacy_setup_content
 
 from .pptoml import (
   pptoml,
+  project,
+  pyproj,
+  pyproj_dist,
+  pyproj_dist_source,
+  pyproj_dist_binary,
+  pyproj_targets,
   # NOTE: deprecated
-  pyproj_meson )
+  pyproj,
+  pyproj_meson,
+  pyproj_targets)
 
 from .builder import (
   Builder )
@@ -67,7 +55,7 @@ from .builder import (
 from .dist_file import (
   dist_copy )
 
-#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+#===============================================================================
 class PyProjBase:
   """Minimal build system for a Python project
 
@@ -76,34 +64,61 @@ class PyProjBase:
 
   Parameters
   ----------
-  root : str | pathlib.Path
+  root :
     Path to the root project directory containing 'pyproject.toml'.
-  logger : logging.Logger
+  logger :
     Parent logger to use when processing project.
 
   """
+
+  root: Path
+  logger: Logger
+  editable: bool
+  pptoml_file: Path
+  pptoml_checksum: tuple[str, int]
+  commit: str
+  env_pkgs: list[str]
+
   #-----------------------------------------------------------------------------
   def __init__( self, *,
-    root,
-    config_settings = None,
-    logger = None ):
+    root: Path,
+    config_settings: dict|None = None,
+    logger: Logger|None = None,
+    editable: bool = False):
 
-    self.logger = logger or logging.getLogger( __name__ )
+    root = resolve(Path(root))
 
-    self.root = Path(root).resolve()
+    self.root = root
+    self.logger = logger or getLogger( __name__ )
+    self.editable = editable
 
     self.pptoml_file = self.root / 'pyproject.toml'
 
     with open( self.pptoml_file, 'rb' ) as fp:
       src = fp.read()
+      self.pptoml_checksum = hash_sha256(src)
+
       src = src.decode( 'utf-8', errors = 'replace' )
       self._pptoml = tomli.loads( src )
 
     with validating(root = self._pptoml, file = self.pptoml_file):
       self._pptoml = pptoml(self._pptoml)
 
+      with validating(key = 'tool'):
+        if 'tool' not in self.pptoml:
+          # TODO: !!!
+          raise RequiredValueError("tool.pyproj is required for backend")
+
+        with validating(key = 'pyproj'):
+          if 'pyproj' not in self.pptoml.tool:
+            raise RequiredValueError("tool.pyproj is required for backend")
+
+      if self.project.dynamic and 'prep' not in self.pyproj:
+        raise RequiredValueError("tool.pyproj.prep is required to resolve project.dynamic")
+
     #...........................................................................
     # construct a validator from the tool.pyproj.config table
+    # NOTE: only really used in the event that "config_settings" are passed to the backend
     config_default = dict()
 
     for k,v in self.pyproj.config.items():
@@ -116,14 +131,12 @@ class PyProjBase:
       else:
         config_default[k] = valid(v, type(v))
 
-    class valid_config(valid_dict):
-      _allow_keys = list()
-      _default = config_default
+    class valid_config_settings(valid_dict):
+      allow_keys = list()
+      default = config_default
 
     with validating( key = 'config_settings' ):
-      _config_settings = valid_config(config_settings or dict())
-
-      self.pyproj.config = _config_settings
+      self._config_settings = valid_config_settings(config_settings or dict())
 
     #...........................................................................
     self.build_backend = mapget( self.pptoml,
@@ -149,8 +162,8 @@ class PyProjBase:
     self.prep()
 
     with validating(
-      key = 'project', 
-      root = self._pptoml, 
+      key = 'project',
+      root = self._pptoml,
       file = self.pptoml_file):
 
       self.pkg_info = PkgInfo(
@@ -160,39 +173,62 @@ class PyProjBase:
     # Update logger once package info is created
     self.logger = self.logger.getChild( f"['{self.pkg_info.name_normed}']" )
 
+    # check if building from git repo
+    commit = ''
+
+    if (root/'.git').is_dir():
+      commit = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).decode('utf-8').strip()
+
+    self.commit = commit
+
+    # inspect/record the environment for installed packages
+    self.env_pkgs = sorted(set([
+      f"{pkg.metadata['Name']}=={pkg.metadata['Version']}"
+      for pkg in metadata.Distribution.discover()]))
+
+    # ensure that essential files will be in the source distribution
+    essential = [
+      Path('pyproject.toml'),
+      self.project.get('readme', {}).get('file'),
+      self.project.get('license', {}).get('file')]
+
+    for file in essential:
+      if not (file is None or any(c.src == file for c in self.source.copy)):
+        self.source.copy.append(file)
+
   #-----------------------------------------------------------------------------
   @property
-  def pptoml(self):
+  def pptoml(self) -> pptoml:
     """pptoml : Parsed and validated pyproject.toml document
     """
     return self._pptoml
 
   #-----------------------------------------------------------------------------
   @property
-  def project(self):
+  def project(self) -> project:
     """:class:`partis.pyproj.pptoml.project`
     """
     return self._pptoml.project
 
   #-----------------------------------------------------------------------------
   @property
-  def pyproj(self):
-    """:class:`partis.pyproj.pptoml.pyproj`
-    """
+  def pyproj(self) -> pyproj:
     return self._pptoml.tool.pyproj
 
   #-----------------------------------------------------------------------------
   @property
-  def config(self):
-    """:class:`partis.pyproj.pptoml.pyproj_config`
+  def config_settings(self):
+    """Config settings passed to backend, or defaults from ``pyproj.config``
     """
-    return self._pptoml.tool.pyproj.config
+    return self._config_settings
+
+  #-----------------------------------------------------------------------------
+  # alias for backward compatibility
+  config = config_settings
 
   #-----------------------------------------------------------------------------
   @property
-  def targets(self):
-    """
-    """
+  def targets(self) -> pyproj_targets:
     return self._pptoml.tool.pyproj.targets
 
   #-----------------------------------------------------------------------------
@@ -201,10 +237,13 @@ class PyProjBase:
     """:class:`partis.pyproj.pptoml.pyproj_meson`
 
     .. deprecated:: 0.1.0
-      Use :attr:`PyProjBase.targets` to access all build targets. 
+      Use :attr:`PyProjBase.targets` to access all build targets.
       These are no longer restricted to meson, but this attribute kept for backward
       compatability.
 
+      Inplace changes to the returned object are not propagated back to the target
+      configuration.
+
     """
     targets = self._pptoml.tool.pyproj.targets
 
@@ -213,31 +252,27 @@ class PyProjBase:
 
     meson = dict(targets[0])
     meson.pop('entry')
+    meson.pop('work_dir')
+    meson.pop('env')
+    meson.pop('exclusive')
     meson['compile'] = meson.pop('enabled')
     return pyproj_meson(meson)
 
   #-----------------------------------------------------------------------------
   @property
-  def dist(self):
-    """:class:`partis.pyproj.pptoml.pyproj_dist`
-    """
+  def dist(self) -> pyproj_dist:
     return self._pptoml.tool.pyproj.dist
 
   #-----------------------------------------------------------------------------
   @property
-  def source(self):
-    """:class:`partis.pyproj.pptoml.pyproj_dist_source`
-    """
+  def source(self) -> pyproj_dist_source:
     return self._pptoml.tool.pyproj.dist.source
 
   #-----------------------------------------------------------------------------
   @property
-  def binary(self):
-    """:class:`partis.pyproj.pptoml.pyproj_dist_binary`
-    """
+  def binary(self) -> pyproj_dist_binary:
     return self._pptoml.tool.pyproj.dist.binary
 
-
   #-----------------------------------------------------------------------------
   @property
   def add_legacy_setup(self):
@@ -247,9 +282,7 @@ class PyProjBase:
 
   #-----------------------------------------------------------------------------
   @property
-  def build_requires(self):
-    """set[:class:`PkgInfoReq`]
-    """
+  def build_requires(self) -> set[PkgInfoReq]:
     return self._build_requires
 
   #-----------------------------------------------------------------------------
@@ -282,14 +315,10 @@ class PyProjBase:
     project = deepcopy(self.project)
     dynamic = project.dynamic
 
-
-    if dynamic and 'prep' not in self.pyproj:
-      raise ValidationError(f"tool.pyproj.prep is required to resolve project.dynamic")
-
     self.prep_entrypoint(
-      name = f"tool.pyproj.prep",
+      name = "tool.pyproj.prep",
       obj = self.pyproj,
-      logger = self.logger.getChild( f"prep" ) )
+      logger = self.logger.getChild("prep") )
 
     # NOTE: check that any dynamic meta-data is defined after prep
     for k in dynamic:
@@ -317,9 +346,9 @@ class PyProjBase:
     """
 
     self.prep_entrypoint(
-      name = f"tool.pyproj.dist.prep",
+      name = "tool.pyproj.dist.prep",
       obj = self.dist,
-      logger = self.logger.getChild( f"dist.prep" ) )
+      logger = self.logger.getChild("dist.prep") )
 
 
   #-----------------------------------------------------------------------------
@@ -328,9 +357,9 @@ class PyProjBase:
     """
 
     self.prep_entrypoint(
-      name = f"tool.pyproj.dist.source.prep",
+      name = "tool.pyproj.dist.source.prep",
       obj = self.dist.source,
-      logger = self.logger.getChild( f"dist.source.prep" ) )
+      logger = self.logger.getChild("dist.source.prep") )
 
   #-----------------------------------------------------------------------------
   def dist_source_copy( self, *, dist ):
@@ -342,36 +371,40 @@ class PyProjBase:
       Builder used to write out source distribution files
     """
 
-    with validating( key = f'tool.pyproj.dist.source' ):
+    with validating( key = 'tool.pyproj.dist.source'):
       dist_copy(
         base_path = dist.named_dirs['root'],
-        include = self.source.copy,
+        copy_items = self.source.copy,
         ignore = self.dist.ignore + self.source.ignore,
         dist = dist,
         root = self.root,
         logger = self.logger )
 
       if self.add_legacy_setup:
-        with validating( key = f'add_legacy_setup' ):
+        with validating(key = 'add_legacy_setup'):
 
-          self.logger.info(f"generating legacy 'setup.py'")
+          self.logger.info("generating legacy 'setup.py'")
           legacy_setup_content( self, dist )
 
   #-----------------------------------------------------------------------------
-  def dist_binary_prep( self ):
+  def dist_binary_prep( self, incremental: bool = False ):
     """Prepares project files for a binary distribution
     """
 
-    with Builder(
+    builder = Builder(
       pyproj = self,
       root = self.root,
       targets = self.targets,
-      logger = self.logger.getChild( f"targets" ) ):
+      logger = self.logger.getChild("targets"),
+      editable = self.editable)
+
+    with builder:
+      builder.build_targets()
 
       self.prep_entrypoint(
-        name = f"tool.pyproj.dist.binary.prep",
+        name = "tool.pyproj.dist.binary.prep",
         obj = self.binary,
-        logger = self.logger.getChild( f"dist.binary.prep" ) )
+        logger = self.logger.getChild("dist.binary.prep") )
 
     self.logger.debug(f"Compatibility tags after dist.binary.prep: {self.binary.compat_tags}")
 
@@ -386,16 +419,17 @@ class PyProjBase:
     """
 
 
-    with validating( key = f'tool.pyproj.dist.binary' ):
+    with validating(key = 'tool.pyproj.dist.binary'):
       ignore = self.dist.ignore + self.dist.binary.ignore
 
       dist_copy(
         base_path = dist.named_dirs['root'],
-        include = self.binary.copy,
+        copy_items = self.binary.copy,
         ignore = ignore,
         dist = dist,
         root = self.root,
-        logger = self.logger )
+        logger = self.logger,
+        follow_symlinks = True)
 
       data_scheme = [
         'data',
@@ -415,8 +449,9 @@ class PyProjBase:
           with validating( key = k ):
             dist_copy(
               base_path = dist.named_dirs[k],
-              include = _include,
+              copy_items = _include,
               ignore = _ignore,
               dist = dist,
               root = self.root,
-              logger = self.logger )
+              logger = self.logger,
+              follow_symlinks = True)
